﻿using System.Collections.Generic;
using System.Threading;
using Unity.Mathematics;
using UnityEngine;

public class TerrainGenThread
{
    public Vector2 InitialCrawlPosition { get; private set; }
    public Vector2Int GridPosition { get; private set; }
    private List<Octave> octaves;
    public int Samples { get; private set; }
    public TerrainData TerrainData { get; private set; }

    private float[,] heights;

    private float heightScale = 1;

    public bool Ready { get; private set; }

    public TerrainGenThread(Vector2 initialCrawlPos, Vector2Int gridPos, List<Octave> octaves, int samples, float heightScale, TerrainData data)
    {
        InitialCrawlPosition = initialCrawlPos;
        GridPosition = gridPos;
        Samples = samples;
        this.octaves = octaves;
        TerrainData = data;
        this.heightScale = heightScale;
    }

    /// <summary>
    /// Gets the generated heights.
    /// Throws InvalidOperationExcepion if the heights
    /// aren't built yet.
    /// </summary>
    /// <returns></returns>
    public float[,] GetHeights()
    {
        if (Ready)
            return heights;
        else
            throw new System.InvalidOperationException("Heights not built yet.");
    }

    public void Generate()
    {
        heights = new float[Samples, Samples];

        foreach (var octave in octaves)
        {
            for (int x = 0; x < Samples; x++)
            {
                for (int y = 0; y < Samples; y++)
                {
                    float2 crawlPos = new((x + InitialCrawlPosition.x * (Samples - 1)) * octave.frequency,
                        (y + InitialCrawlPosition.y * (Samples - 1)) * octave.frequency);
                    
                    float addition;

                    switch (octave.type)
                    {
                        default:
                        case NoiseType.MathfPerlin:
                            addition = Mathf.PerlinNoise(crawlPos.x, crawlPos.y);
                            break;
                        case NoiseType.MathematicsPerlin:
                            addition = noise.cnoise(crawlPos) / 2.3f + 0.5f;
                            break;
                        case NoiseType.Simplex:
                            addition = noise.snoise(crawlPos) / 4.6f + 0.5f;
                            break;
                        case NoiseType.Celluar:
                            addition = noise.cellular(crawlPos).x / 2.3f + 0.5f;
                            break;
                    }
                    heights[y, x] += addition * octave.amplitude * heightScale;
                }
            }
        }

        Ready = true;
    }
}