using System.Threading;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Unity.Mathematics;
using System.Collections.Concurrent;

[System.Serializable]
public struct Octave
{
    public float amplitude;
    public float frequency;
    public NoiseType type;
}

[System.Serializable]
public enum NoiseType
{
    MathfPerlin,
    MathematicsPerlin,
    Simplex,
    Celluar
}

public class TerrainGenerator : MonoBehaviour
{
    [Header("User settings")]
    public List<Octave> octaves;

    [Tooltip("The seed of the generation.")]
    public Vector2 crawlStartPosition;

    public bool generateNewSeed = true;

    public bool refresh;

    [Tooltip("How fine grain the generated terrain will be. " +
        "Larger values = more precise. Recommended value is 513.")]
    public int samples = 513;

    [Tooltip("How tall the terrain is. May be changed to allow for higher terrains.")]
    public float terrainHeight = 600;

    [Header("Autogenerated values")]
    [Tooltip("Which terrain grid are we on right now?")]
    public Vector2Int currentGridPosition;

    [Tooltip("Dictionary of previously loaded terrains.")]
    public Dictionary<Vector2Int, TerrainData> previouslyLoaded = new();

    ///// <summary>
    ///// Threads used for generating the neighboring and current terrains.
    ///// </summary>
    //private ConcurrentBag<Thread> threads = new();

    private ConcurrentQueue<TerrainGenThread> generators = new();

    public int bagCount = 0;

    /// <summary>
    /// Apply the terrain that was generated by genThread.
    /// </summary>
    /// <param name="genThread">The terrain gen thread that has finished 
    /// generating its terrain heights.</param>
    private void ApplyGeneratedTerrain(TerrainGenThread genThread)
    {
        genThread.TerrainData.SetHeights(0, 0, genThread.GetHeights());

        GameObject obj = Terrain.CreateTerrainGameObject(genThread.TerrainData);
        obj.name = genThread.InitialCrawlPosition.ToString() + "Terrain";
        obj.transform.SetParent(transform, false);
        Vector3 offset = new Vector3(genThread.GridPosition.x, 0, genThread.GridPosition.y) *
            genThread.TerrainData.size.x;
        obj.transform.position += offset;
    }

    private void GenerateThread(Vector2 crawlPos, Vector2Int gridPos,
         bool forceRefresh)
    {
        if (!previouslyLoaded.ContainsKey(gridPos) || forceRefresh)
        {
            TerrainData data = Instantiate(Resources.Load<TerrainData>("Template"));
            TerrainGenThread genThread = new(crawlPos, gridPos, octaves, samples,
                data.size.y / terrainHeight, data);
            data.size = new(data.size.x, terrainHeight, data.size.z);

            generators.Enqueue(genThread);
            bagCount++;
            ThreadStart starter = new(genThread.Generate);
            Thread thread = new(starter);
            previouslyLoaded[gridPos] = data;
            thread.Start();
        }
    }

    public virtual void GenerateSurronding(Vector2 initCrawlPos, bool forceRefresh = false)
    {
        int count = 0;
        for (int i = -1; i <= 1; i++)
        {
            for (int j = 1; j >= -1; j--)
            {
                GenerateThread(currentGridPosition + initCrawlPos + new Vector2(i, j),
                    currentGridPosition + new Vector2Int(i, j),
                    forceRefresh);
                count++;
            }
        }

        StartCoroutine(CheckThreads());
    }

    /// <summary>
    /// Wait until all threads are done.
    /// </summary>
    /// <returns></returns>
    private IEnumerator CheckThreads()
    {
        yield return new WaitForSecondsRealtime(1);

        //bool busy = true;

        //while (busy)
        //{
        //    busy = false;

        //    foreach (var thread in generationThreads)
        //    {
        //        // Wait for all threads to finish.
        //        if (thread != null && thread.IsAlive)
        //        {
        //            busy = true;
        //            yield return new WaitForEndOfFrame();
        //            break;
        //        }
        //    }
        //}

        while (!generators.IsEmpty)
        {
            if (generators.TryPeek(out TerrainGenThread tGen))
            {
                if (tGen.Ready)
                {
                    generators.TryDequeue(out _);
                    ApplyGeneratedTerrain(tGen);
                    bagCount--;
                }

                yield return new WaitForEndOfFrame();
            }
        }
    }

    private void Update()
    {
        if (refresh)
        {
            refresh = false;

            foreach (Transform child in transform)
                Destroy(child.gameObject);

            GenerateSurronding(crawlStartPosition, true);
        }
    }
}
